import Board
import BotHelper
import Robot
import Client

import Data.Array.IArray
import List
import Maybe
import Monad
import System

import IOExts
import Data.FiniteMap

import Distance

type BotPass = BotInfo -> IO (Maybe (Bid,Command))

bot :: AI_Robot ()
bot () bi = bot_move bi 

run_passes [] _ = error "No pass, impossible"
run_passes ((p,t):rest) bi = do
  v <- p bi
  case v of
    Nothing -> run_passes rest bi
    Just z -> return (z,t)

bot_move bi = do
  let passes = [(killer_move,  "KILLER"),
		(deliver_here, "deliver here"),
		(near_home,    "near_home"),
		(deliver,      "deliver"),
		(find_packages,"find package"),
		(load_robot,   "load robot"),
		(not_visited,  "not_visited"),
		(panic,        "panic")]
  ((b,c),t) <- run_passes passes bi
  let money = fuel bi
  if b > money
     then return ((money,c,t),())
     else return ((b,c,t),())

create_stand_bid bi = do
  let money = fuel bi
  let other_robots = [sq | (_,(sq,_)) <- robots bi]
  let around = walk (current_sq bi) `intersect` other_robots
  in_danger <- liftM or $ sequence [square_is Water (board bi) behind
				    | sq <- around
				    , let d = dir sq (current_sq bi)
				    , let behind = next_square sq d]
  case  True of 
   _ | in_danger         -> return $ money `div` (length other_robots)
     | not $ null around -> return $ money `div` (8 * length other_robots)
     | otherwise         -> return 1

create_move_bid bi pos = do
  let d = dir (current_sq bi) pos
  simple_water_kill <- square_is Water (board bi) (next_square pos d)

  let other_robots = [sq | (_,(sq,_)) <- robots bi]
  let places_they_can_be_in = concat [walk sq | sq <- other_robots]
  let in_my_way = pos `elem` other_robots
  let maybe_in_my_way = pos `elem` places_they_can_be_in
  let have_cargo = not $ null (carry bi)
  let simple_kill = in_my_way && simple_water_kill
  let around = walk (current_sq bi) `intersect` other_robots
  in_danger <- liftM or $ sequence [square_is Water (board bi) behind
				    | sq <- around
				    , let d = dir (current_sq bi) sq
				    , let behind = next_square (current_sq bi) (opposite_dir d)]
  let money = fuel bi
  case True of 
   _ | in_danger   -> return $ money `div` (4 * length other_robots)
     | simple_kill -> return $ money `div` (8 * length other_robots)
     | in_my_way   -> return $ money `div` (16 * length other_robots)
     | otherwise   -> return 1

find_packages :: BotPass
find_packages bi = do
  let known_places = eltsFM (pkg_place bi)
  let dist = dist_get (dist_from_me bi)
  known2 <- sequence [dist sq >>= \d -> return (d,sq) | sq <- (nub known_places)]
  let neareast_known_place = snd $ minimum known2
  let move = move_command (current_sq bi)
  case True of
   _ | notNull packages_on_my_square -> return $ Nothing -- we have already found
     | notNull known_places          -> do dest <- goto bi neareast_known_place
					   bid <- create_move_bid bi dest
					   return $ Just (bid, move dest)
     | notNull (home bi)             -> do Just sq <- nearest bi (home bi)
					   dest <- goto bi sq
					   bid <- create_move_bid bi dest
					   return $ Just (bid, move dest)
     | otherwise                     -> return Nothing
    
    where notNull = not . null
	  packages_on_my_square = packages bi

load_robot :: BotPass
load_robot bi 
  | null (packages bi) = return $ Nothing
  | otherwise = do c <- fill_up bi
		   bid <- create_stand_bid bi
		   return $ Just (bid, c)

deliver :: BotPass
deliver bi
  | null (carry bi) = return Nothing
  | otherwise = -- drop here is another pass, find nearest destination
                let v = \sq -> unsafePerformIO (dist_get (dist_from_me bi) sq)
		    (_,sq) = minimum [ (d,sq)
				       | (p,(sq,_)) <- carry bi
				       , let d = v sq ]
		in do dest <- goto bi sq
		      let move = move_command (current_sq bi)
		      bid <- create_move_bid bi dest
		      return $ Just (bid, move dest)

-- if I stand next to a robot and he has water on the other side, then push him
killer_move :: BotPass
killer_move bi =
  let other_robots = [sq | (_,(sq,_)) <- robots bi]
      around = walk (current_sq bi) `intersect` other_robots
      money = fuel bi
  in do in_danger <- do let cand = [ (sq,next) 
				     | sq <- around,
				     let d = dir (current_sq bi) sq,
				     let next = next_square sq d]
			xs <- filterM (\(q,_) -> can_go_to (board bi) q) cand
			ys <- filterM (\(_,n) -> square_is Water (board bi) n) xs
			return $ map fst ys

        case in_danger of
          [] -> return Nothing
	  (x:xs) -> do let f = fuel bi
		       let m = Move (dir (current_sq bi) x)
		       let bid = money `div` (2 * length other_robots)
		       return $ Just (bid, m)

near_home :: BotPass
near_home bi = do
  let dist = dist_get (dist_from_me bi)
  let move = move_command (current_sq bi)
  let hs = home bi
  xs <- sequence [dist h >>= \x -> return (x,h) | h <- hs]
  case xs of
    [] -> return Nothing
    xs -> do let (d,sq) = minimum xs
             if d < 10 * 10 -- 10 is approx on square
		then  do dest <- goto bi sq
			 bid <- create_move_bid bi dest
			 return $ Just (bid, move dest)
		else return Nothing

deliver_here :: BotPass
deliver_here bi
  | null (carry bi) = return Nothing
  | otherwise = 
      let droppable = [p |(p,(sq,_)) <- carry bi, sq == current_sq bi]
      in case droppable of
	     [] -> -- so can't drop here
		   return Nothing
             xs -> do bid <- create_stand_bid bi
	              return $ Just (bid, Drop xs)

not_visited :: BotPass
not_visited bi = do
  let (Board w h _) = board bi
  let here = current_sq bi
  let filter_not_visited = filterM (\x -> liftM not (board_visited (board bi) x))
  sq <- filterM (can_go_to (board bi)) (walk here)
  sq <- filterM (can_go_to (board bi)) (concatMap walk sq)
  sq <- filterM (can_go_to (board bi)) (concatMap walk sq)
  sq <- return $ nub sq		       
  xs <- filter_not_visited sq
  let dist = dist_get (dist_from_me bi)
  let move = move_command (current_sq bi)
  ys <- sequence [dist sq >>= (\y -> return (y,sq)) | sq <- xs]
  case ys of
    [] -> -- we have tried to find a square near, now take any square
          do  xs <- filter_not_visited [(x,y) | x <- [1..w], y <- [1..h]]
	      ys <- filterM (reachable bi) xs
	      case ys of
	        [] -> return Nothing
		(y:_) -> do dest <- goto bi y
			    bid <- create_move_bid bi dest
			    return $ Just (bid, move dest)
    xs -> do let (_,sq) = minimum xs
	     dest <- goto bi sq
	     bid <- create_move_bid bi dest
	     return $ Just (bid, move dest)

panic bi = return $ Just (1, Drop [])

main = start_bot bot ()
test = start_bot_server "localhost" 5000 bot ()
