module BotHelper where

import Array
import List
import IOExts
import Monad
import Random

import Ansi
import Board
import Client
import Distance
import Robot

walk (x,y) = [(x+1,y), (x-1,y), (x,y+1), (x,y-1)]

goto :: BotInfo -> Square -> IO Square
goto bi pos = do
  d <- new_distance (board bi)
  update_distance_to d pos
  dists <- do sqs <- filterM (can_go_to (board bi)) $ walk (current_sq bi)	      
	      sequence [do x <- dist_get d sq
		           first <- dist_get (dist_from_me bi) sq
		           rand <- randomIO -- for equal dists we chose a random path
			   -- putStrLn $ show (first,x, sq) ++ clear_eol
		           return ((x+first, rand :: Int),sq)
			| sq <- sqs]
  free_distance d
  return $ (snd $ minimum dists)
--  return $ dir_command (current_sq bi) (snd $ minimum dists)

move_command :: Square -> Square -> Command
move_command from@(x1,y1) to@(x2,y2) 
    | x1 == x2 && y1+1 == y2 = Move North
    | x1 == x2 && y1-1 == y2 = Move South
    | x1+1 == x2 && y1 == y2 = Move East
    | x1-1 == x2 && y1 == y2 = Move West
    | otherwise = Drop [] -- should not happen

dir :: Square -> Square -> Dir
dir from@(x1,y1) to@(x2,y2) 
    | x1 == x2 && y1+1 == y2 = North
    | x1 == x2 && y1-1 == y2 = South
    | x1+1 == x2 && y1 == y2 = East
    | x1-1 == x2 && y1 == y2 = West
    | otherwise              = North -- should not happend, but it is a contest. morphy rules.

opposite_dir North = South
opposite_dir South = North
opposite_dir West = East
opposite_dir East = West


next_square :: Square -> Dir -> Square
next_square (x,y) North = (x,y+1)
next_square (x,y) South = (x,y-1)
next_square (x,y) West  = (x-1,y)
next_square (x,y) East  = (x+1,y)

reachable :: BotInfo -> Square -> IO Bool
reachable bi sq = do
  let dist = dist_get (dist_from_me bi)
  d <- dist sq
  return $ d < 1000000000

nearest :: BotInfo -> [Square] -> IO (Maybe Square)
nearest bi sqs = do
  let dist = dist_get (dist_from_me bi)
  xs <- sequence [dist sq >>= (\d -> return (d,sq)) | sq <- sqs]
  case xs of
    [] -> return $ Nothing
    _  -> return $ Just $ snd $ minimum xs

--good_package bi =
--  let [ (id,dest,w) <- packages bi, w <=  ]

-- 1. Pick the package that is the closes to deliver
-- 2. Pick up the package that is closest to deliver from the
--    last drop spot
-- 3. continue with 2
 
p_dest   (id,(sq,w)) = sq
p_weight (id,(sq,w)) = w

fill_up :: BotInfo -> IO Command
fill_up bi =
  let pos = current_sq bi
      free = free_weight bi
      pkgs = packages bi
      fu pos free pkgs =
	  do d <- new_distance (board bi)
	     update_distance_to d pos
	     candidates <- sequence [ do di <- dist_get d (p_dest p)
				         return (di,p)
				      | p <- pkgs
				      , p_weight p <= free]
	     let (d,p) = minimum candidates
	     case candidates of
			[] -> return []
			xs -> do ps <- fu (p_dest p) (free - p_weight p) ((map snd candidates) \\ [p])
			         return (p:ps)
  in do xs <- fu pos free pkgs	 
	return $ Pick (map fst xs)

