module Robot where

import Board
import Package

import IOExts

type RobotID = Int
type MaxWeight = Int
type Money = Int
type Bid = Int

type RobotInfo = (Square, [PackageID])
type Robot = (RobotID, RobotInfo)
type Me = (RobotID, MaxWeight, Money)

{-
* N The robot moved north.
* S The robot moved south.
* E The robot moved east.
* W The robot moved west.
* P id The robot picked up package with id id.
* D id The robot dropped package with id id.
* X num The robots x-coordinate is set to num.
* Y num The robots y-coordinate is set to num. 
-}

data Action
 = N
 |S
 | E
 | W
 | P ! PackageID
 | D ! PackageID
 |X ! CoordX
 |Y ! CoordX
   deriving (Show)

data Dir
 = North
 | East
 | South
 | West
   deriving (Show)

data Command
 = Move Dir
 | Pick [PackageID]
 | Drop [PackageID]
   deriving (Show)

parse_robot_actions :: String -> [ (RobotID, [Action]) ]
parse_robot_actions str = pc (words str)
  where pc [] = []
	pc xs@(('#':_):_) = let (rid,cs,ts) = parse_robot_action xs
			    in ((rid,cs) : pc ts)
	pc _ = trace ("Invalid command: " ++ str) -- error, end message?
	       []

parse_robot_action :: [String] -> (RobotID, [Action], [String])
parse_robot_action (('#':rid):ts) = (read rid, ac, rest)
   where (ac,rest) = parse_actions ts

         next c ts = (c:cs,ts') where (cs,ts') = pc ts

	 pc ("N":ts) = next N ts
	 pc ("S":ts) = next S ts
	 pc ("W":ts) = next W ts
	 pc ("E":ts) = next E ts
	 pc ("P":x:ts) = next (P (read x)) ts
	 pc ("D":x:ts) = next (D (read x)) ts
	 pc ("X":x:ts) = next (X (read x)) ts
	 pc ("Y":x:ts) = next (Y (read x)) ts
	 pc ts = ([],ts)
parse_robot_action q = error $ "parse_robot_action "  ++ show q

parse_actions ts = p ts
   where next c ts = (c:cs,ts') where (cs,ts') = p ts

	 p ("N":ts) = next N ts
	 p ("S":ts) = next S ts
	 p ("W":ts) = next W ts
	 p ("E":ts) = next E ts
	 p ("P":x:ts) = next (P (read x)) ts
	 p ("D":x:ts) = next (D (read x)) ts
	 p ("X":x:ts) = next (X (read x)) ts
	 p ("Y":x:ts) = next (Y (read x)) ts
	 p ts = ([],ts)


command2str :: Command -> String
command2str (Move North) = "Move N"
command2str (Move South) = "Move S"
command2str (Move West)  = "Move W"
command2str (Move East)  = "Move E"
command2str (Pick xs)    = "Pick" ++ concat [" " ++ (show c) | c <- xs]
command2str (Drop xs)    = "Drop" ++ concat [" " ++ (show c) | c <- xs]

bid_cmd2str :: (Bid, Command) -> String
bid_cmd2str (bid, c) = show bid ++ " " ++ command2str c

