#define NO_DISTANCE 1000000000

//////////////////////////////////////
// Simple heap

static void swap (int *heap, int a, int b)
{
  int tmp;

  tmp = heap[a];
  heap[a] = heap[b];
  heap[b] = tmp;
}

static void sift_down (int *dist, int *heap, int *size, int pos)
{
  int k, j;

  k = pos;

  do {
    j = k;

    if ( (2*j) <= *size && dist[heap[2*j]] < dist[heap[k]])
      k = 2*j;

    if ( (2*j) < *size && dist[heap[2*j+1]] < dist[heap[k]])
      k = 2*j+1;

    swap (heap, j, k);
  }
  while (j != k);
}

static void percolate (int *dist, int *heap, int *size, int pos)
{
  int k, j;

  k = pos;

  do {
    j = k;
    if (j > 1 && dist[heap[j / 2]] > dist[heap[k]])
      k = j / 2;
    
    swap (heap, k, j);
  }
  while ( j != k);
}

static void heap_add (int *dist, int *heap, int *size, int pos)
{
  (*size)++;

  heap[*size] = pos;

  percolate (dist, heap, size, *size);
}

static int heap_take (int *dist, int *heap, int *size)
{
  int x;

  x = heap[1];

  heap[1] = heap[*size];

  (*size)--;

  sift_down (dist, heap, size, 1);

  return x;
}

//////////////////////////////////////
// Shortest path
//
// Dijkstras algoritm (almost)
static void clear (int *res, int w, int h)
{
  int i;

  for (i = 0; i < w*h; i++)
    res[i] = NO_DISTANCE;
}

#define POS(x,y) ((x)+(y)*w)

#define WATER 3
#define WALL  2
#define HOME  0

#define VISITED (0x08)
#define ROBOT   (0x10)

void calc_distance (char *board, int *dist, int w, int h, int x, int y, int from)
{
  int *heap = (int *)malloc (w*h*sizeof(int));
  int size = 0, p2;

  x--;
  y--;

  clear (dist, w, h);

  heap_add (dist, heap, &size,  x+y*w);
  dist[x+y*w] = 0;

  while (size > 0) {
    int pos = heap_take (dist, heap, &size);
    int y2[4], x2[4], i;

    x = pos % w;
    y = pos / w;

    x2[0] = x-1; x2[1] = x+1; x2[2] = x;   x2[3] = x;
    y2[0] = y;   y2[1] = y;   y2[2] = y+1; y2[3] = y-1;

    for (i=0; i<4; i++) {
      int v, cost;

      x = x2[i]; y = y2[i];

      if ( x < 0 || x >= w || y < 0 || y >= h)
	continue;

      p2 = POS(x,y);

      v = board[p2] & 0x03;

      cost = 9;
      if (board[from ? p2 : pos] & VISITED)
	cost = 10;

      if (board[from ? p2 : pos] & 0x04)
	cost += 18;

      if (board[from ? p2 : pos] & ROBOT)
	cost += 7;

      if ( v !=  WATER && v != WALL && (dist[pos] + cost) < dist[p2] )
      {
	dist [p2] = dist[pos] + cost;

	heap_add (dist, heap, &size,  p2);
      }
    }
  }

  free (heap);
}

////////////////////////////////////
// Static info about map
void water_info (char *board, int w, int h)
{
  int x,y;
  int y2[4], x2[4], i;

  for (x=0; x<w; x++) {
    for (y=0; y<h; y++) {
      if ( (board[x+y*w] & 0x03) == WATER ) {
	x2[0] = x-1; x2[1] = x+1; x2[2] = x;   x2[3] = x;
	y2[0] = y;   y2[1] = y;   y2[2] = y+1; y2[3] = y-1;
	
	for (i=0; i<4; i++) {	  
	  int a,b;

	  a = x2[i]; b = y2[i];

	  if (a >= 0 && a < w && b >= 0 && b < h) {
	    board[a+b*w] |= 0x4;
	  }
	}
      }
    }
  }
}
